from __future__ import print_function
from __future__ import division
from builtins import str
from builtins import input
from builtins import range
from past.utils import old_div
from builtins import object
#from SWIFr_train import AODE_train
from sklearn import mixture
import os, pickle, math, sys, numpy as np, argparse

sys.path.append(os.getcwd())

class Mix1D(object):
    def __init__(self):
        self.means_ = []
        self.weights_ = []
        self.covariances_ = []

class Stats(object):
    def __init__(self,path):
        if path != '' and path[-1] != '/':
            path += '/'
        file = open(path+'component_stats.txt')
        f = file.read()
        file.close()
        f = f.strip().splitlines()
        self.stats = [line.split('\t')[0] for line in f]
        self.stat2score = {s:-998 for s in self.stats}
    def set_stat(self,stat,value):
        self.stat2score[stat] = value


class AODE(object):
    '''
    Given a directory containing AODE parameters generated by SWIFr_train.py, compute
    posterior probabilities for each classification scenario. Can be run site-by-site
    using --interactive, or can be computed for every SNP in a file using --file. One 
    or the other flag must be provided.
    '''

    def __init__(self,path2trained):

        if path2trained != '' and path2trained[-1] != '/':
            path2trained += '/'
        self.path2trained = path2trained

        file = open(self.path2trained+'component_stats.txt','r')
        f = file.read()
        file.close()
        f = f.strip().splitlines()
        f = [x.strip() for x in f]
        self.statlist = f

        self.num2stat = {i:self.statlist[i] for i in range(len(self.statlist))}
        self.stat2num = {y:x for x,y in list(self.num2stat.items())}

        self.path2AODE = self.path2trained+'AODE_params/'    

        file = open(self.path2trained+'classes.txt','r')
        f = file.read()
        file.close()
        self.scenarios = [x.strip() for x in f.strip().splitlines()]
        print('classes:')
        print(self.scenarios)



        self.JOINTS = [[[] for stat2 in list(self.stat2num.keys())] for stat1 in list(self.stat2num.keys())]
        self.MARGINALS = [[] for stat in list(self.stat2num.keys())]

        for stat1 in self.statlist:
            for stat2 in [x for x in self.statlist if x!= stat1]:
                statnum1 = self.stat2num[stat1]
                statnum2 = self.stat2num[stat2]
                if statnum1 < statnum2:
                    for scenario in self.scenarios:
                        self.JOINTS[statnum1][statnum2].append(self.gmm_fit(stat1,stat2,scenario))

        for stat1 in self.statlist:
            statnum1 = self.stat2num[stat1]
            for scenario in self.scenarios:
                self.MARGINALS[statnum1].append(self.gmm_fit_1D(stat1,scenario))

    
    def aode(self,stats,pi):
        #pi must have length len(scenarios)
        numerators = [[] for i in range(len(self.scenarios))]
        raw_nums = [[] for i in range(len(self.scenarios))]
        denominators = []
        ode_num_dicts = {}
        for i in range(len(self.scenarios)):
            ode_num_dicts[self.scenarios[i]]=dict()
        for keystat in self.statlist:
            ps,nums,denom,raw = self.ode(keystat,stats,pi)
            for i in range(len(self.scenarios)):
                if ps != 'n/a':
                    ode_num_dicts[self.scenarios[i]][keystat]=nums[i]
                    numerators[i].append(nums[i])
                    raw_nums[i].append(raw[i])
                else:
                    ode_num_dicts[self.scenarios[i]][keystat]='-998'
            denominators.append(denom)
        aoderaw = [sum(raw_nums[i]) for i in range(len(self.scenarios))]
        if sum(denominators) > 0:
            aodeposteriors = [float(sum(numerators[i]))/sum(denominators) for i in range(len(self.scenarios))]
            return aodeposteriors, aoderaw, ode_num_dicts
        else:
            return [0 for i in range(len(self.scenarios))], aoderaw, ode_num_dicts

    def ode(self,keystat,stats,pi):
        score = stats.stat2score[keystat]
        if score == -998:
            return 'n/a',0,0,0
        else:
            MARGS = self.MARGINALS[self.stat2num[keystat]]
            Likelihoods = []
            for i in range(len(self.scenarios)):
                M = MARGS[i]
                Likelihoods.append(self.GMM_pdf(M,score))

            for stat in self.statlist:
                if stat != keystat:
                    score2 = stats.stat2score[stat]
                    if score2 != -998:
                        if self.stat2num[keystat] < self.stat2num[stat]:
                            for i in range(len(self.scenarios)):
                                H = self.conditional_GMM(score,2,self.JOINTS[self.stat2num[keystat]][self.stat2num[stat]][i])
                                Likelihoods[i] = Likelihoods[i]*self.GMM_pdf(H,score2)
                        else:
                            for i in range(len(self.scenarios)):
                                H = self.conditional_GMM(score,1,self.JOINTS[self.stat2num[stat]][self.stat2num[keystat]][i])
                                Likelihoods[i] = Likelihoods[i]*self.GMM_pdf(H,score2)
            numerators = [0 for i in range(len(self.scenarios))]
            for i in range(len(self.scenarios)):
                numerators[i] = float(pi[i]*Likelihoods[i])
            even_raw_aode = [0 for i in range(len(self.scenarios))]
            for i in range(len(self.scenarios)):
                even_raw_aode[i] = float(0.5*Likelihoods[i])
            denominator = sum(numerators)
            if denominator == 0:                
                return [0 for i in range(len(numerators))],numerators,denominator, Likelihoods
            else:
                posteriors = [float(numerators[i])/denominator for i in range(len(numerators))]
                return posteriors, numerators, denominator, even_raw_aode


    
    def GMM_pdf(self,G,x):
        w = G.weights_
        mu = G.means_
        C = G.covariances_
        pdf = 0
        for i in range(len(w)):
            pdf += w[i]*self.normpdf(x,mu[i][0],math.sqrt(C[i][0]))
        return pdf


    def normpdf(self,x,mu,sigma):
        u = float(x-mu)/sigma
        y = (old_div(1,(math.sqrt(2*math.pi)*abs(sigma))))*math.exp(old_div(-u*u,2))
        return y

    def calc_srs(self,raw_probs):
        log_probs=[]
        for i in raw_probs:
            if i>0:
                log_probs.append(math.log10(i))
            else:
                log_probs.append(float('-inf'))
        return max(log_probs)
    
    def conditional_GMM(self,condval,keystat,G):
        #keystat = 1 if want stat1|stat2, keystat = 2 if want stat2|stat1
        H = Mix1D()

        for i in range(len(G.weights_)):
            sigma1 = math.sqrt(G.covariances_[i][0][0])
            sigma2 = math.sqrt(G.covariances_[i][1][1])
            ro = float(G.covariances_[i][0][1])/(sigma1*sigma2)
            mu1 = G.means_[i][0]
            mu2 = G.means_[i][1]
            if keystat == 1:
                H.weights_.append(G.weights_[i])
                H.means_.append([mu1 + float(sigma1*ro*(condval-mu2))/sigma2])
                H.covariances_.append([(1-ro**2)*sigma1**2])

            elif keystat == 2:
                H.weights_.append(G.weights_[i])
                H.means_.append([mu2 + float(sigma2*ro*(condval-mu1))/sigma1])
                H.covariances_.append([(1-ro**2)*sigma2**2])
        return H

    def gmm_fit(self,stat1,stat2,scenario):
        sys.path.append(os.getcwd())
        G = pickle.load(open(self.path2AODE+stat1+'_'+stat2+'_'+scenario+'_GMMparams.p','rb'))
        return G

    def gmm_fit_1D(self,stat,scenario):
        G = pickle.load(open(self.path2AODE+stat+'_'+scenario+'_1D_GMMparams.p','rb'))
        return G

    def calculate_on_file(self,filename,pivec,outfile,A,nb,ode_diagnostics):
        file = open(filename,'r')
        f = file.read()
        file.close()
        out = open(outfile,'w')
        #out = open(filename.replace('.txt','')+'_classified','w')
        f = f.strip().splitlines()
        header = f[0]
        H = header.strip().split('\t')
        newheader = header+'\t'
        for s in self.scenarios:
            newheader += 'pi_'+s+'\t'
        for s in self.scenarios:
            newheader += 'P('+s+')'+'\t'
        newheader += 'SRS'+'\t'
        if ode_diagnostics:
            for s in self.scenarios:
                newheader += 'RawAODE_P('+s+')'+'\t'
                for stat in self.statlist:
                    newheader += stat + '_ODE('+s+')' + '\t'
        if nb:
            for s in self.scenarios:
                newheader += 'NB_P('+s+')'+'\t'
        newheader = newheader.strip()
        out.write(newheader+'\n')
        stat2index = {}
        for stat in self.statlist:
            stat2index[stat] = H.index(stat)
        f = f[1:]
        for line in f:
            L = line.strip().split('\t')
            S = Stats(self.path2trained)
            for stat in S.stats:
                S.stat2score[stat] = float(L[stat2index[stat]])
            scenario_probs, raw_probs, odes = A.aode(S,pivec)
            outtext = line+'\t'
            for i in range(len(pivec)):
                outtext += str(pivec[i])+'\t'
            for i in range(len(self.scenarios)):
                outtext += str(scenario_probs[i])+'\t'
            outtext += str(A.calc_srs(raw_probs))+'\t'
            if ode_diagnostics:
                for i in range(len(self.scenarios)):
                    outtext += str(raw_probs[i])+'\t'
                    for stat in self.statlist:
                        outtext += str(odes[self.scenarios[i]][stat])+'\t'
            if nb:
                nb_scores, nums, denoms = A.naive_bayes(S,pivec)
                for i in range(len(self.scenarios)):
                    outtext += str(nb_scores[i])+'\t'
            outtext = outtext.strip()
            out.write(outtext+'\n')
        out.close()


    def naive_bayes(self,stats,pi):
        #pi must have length len(scenarios)
        Likelihoods = [1 for i in range(len(self.scenarios))] #product of P(statistic value | scenario) (for each scenario, a product across all statistics)
        for stat in self.statlist:
            score = stats.stat2score[stat] #value of statistic
            MARGS = self.MARGINALS[self.stat2num[stat]] #1-D distributions learned from SWIFr
            for i in range(len(self.scenarios)):
                M = MARGS[i] #distribution for scenario and stat
                Likelihoods[i] = Likelihoods[i]*self.GMM_pdf(M,score)
        numerators = [pi[i]*Likelihoods[i] for i in range(len(self.scenarios))] #multiply by prior scenario probs
        denominator = sum(numerators) #normalization
        # print(posteriors)
        # print(numerators)
        # print(denominator)
        if denominator == 0:
            return [0 for i in range(len(numerators))],numerators, denominator
        else:
            posteriors = [float(numerators[i])/denominator for i in range(len(numerators))]
            return posteriors, numerators, denominator




# if __name__ == '__main__':
def main():

    parser = argparse.ArgumentParser()
    parser.add_argument('--path2trained',action='store',dest='path2trained',default='')
    parser.add_argument('--interactive',action='store_true',dest='interactive',default=False)
    parser.add_argument('--file',action='store',dest='filename') #use instead of interactive mode to work on a whole file
    parser.add_argument('--pi',action='store',nargs='+',default=['0.99999','0.00001']) #can use with either mode
    parser.add_argument('--outfile',action='store',default='')
    parser.add_argument('--nb',action='store_true',dest='nb',default=False) #run naive bayes
    parser.add_argument('--ode',action='store_true',dest='ode',default=False) #output non-normalized AODE output and inidividual ODE scores
    args = parser.parse_args()
     
    if not args.interactive and not args.filename:
        print("Error: SWIF(r) must be run either with an input file using --file or in " \
             + "interactive mode using --interactive.")
        print("Please try running the program again.")
    
    else:
        A = AODE(args.path2trained)
    
        pivec = [float(x) for x in args.pi]
    
        if round(sum(pivec),10) != round(1,10):
            print('Error: prior values should add to 1')
        else:
            if len(pivec) != len(A.scenarios):
                print('Error: need '+str(len(A.scenarios))+' values for pi')
    
            else:
                print('Using priors:')
                for i in range(len(A.scenarios)):
                    print('p('+A.scenarios[i]+') = '+str(pivec[i]))
                if args.interactive:
                    S = Stats(args.path2trained)
                    for stat in S.stats:
                        response = eval(input("Value for "+stat+": "))
                        S.stat2score[stat] = float(response)
                
                    scenario_probs = A.aode(S,pivec)
                    for i in range(len(A.scenarios)):
                        print('Probability of '+A.scenarios[i]+': '+str(scenario_probs[i]))
                else:
                    if args.outfile == '':
                        args.outfile = args.filename.replace('.txt','')+'_classified'
                    A.calculate_on_file(args.filename,pivec,args.outfile, A, args.nb, args.ode)






from sklearn.isotonic import IsotonicRegression as IR
import pickle, os
import numpy as np
import matplotlib
matplotlib.use('PS')
from matplotlib import pyplot as plt

''' take in a tab-delimited file where first column is true values (0=neutral, 1=sweep), and second column is un-calibrated probabilities; as of now, only binary classification is supported'''
class SmoothedIsotonicCalibration:

	def __init__(self):
		if not os.path.isdir('calibration/'):
			os.system('mkdir calibration/')
		self.numbins = 10

	def apply_calibration(self,frac1,filename):
		file = open(filename,'r')
		f = file.read()
		file.close()
		outfilename = os.path.split(filename)[1]
		outfilename = outfilename.replace('.txt','')+'_calibrated'

		f = f.strip().splitlines()
		header = f[0]
		f = f[1:]
		values = []
		header = header.strip().split('\t')
		valix = header.index('uncalibrated')
		oldlines = []
		for line in f:
			oldlines.append(line)
			line = line.strip().split('\t')
			values.append(float(line[valix]))

		M = pickle.load(open('calibration/M_frac1='+str(frac1)+'.p','rb'))
		V = pickle.load(open('calibration/V_frac1='+str(frac1)+'.p','rb'))
		calibrated = self.second_order_correction(values,M,V)

		out = open('calibration/'+outfilename,'w')
		out.write('\t'.join([x for x in header])+'\t'+'calibrated\n')
		for i in range(len(calibrated)):
			out.write(oldlines[i].strip()+'\t'+str(calibrated[i])+'\n')
		out.close()


	def learn_calibration(self,frac1,filename):
		truth,uncalibrated = self.read_file(filename)
		xtrain,ytrain = self.plot_uncalibrated_reliability(truth,uncalibrated,frac1)
		ir = self.isotonic_calibration(xtrain,ytrain)
		pickle.dump(ir,open('calibration/IR_frac1='+str(frac1)+'.p','wb'))
		M,V = self.interpolate_isotonic(ir)
		pickle.dump(M,open('calibration/M_frac1='+str(frac1)+'.p','wb'))
		pickle.dump(V,open('calibration/V_frac1='+str(frac1)+'.p','wb'))
		self.plot_calibrated_reliability(truth,uncalibrated,frac1,ir,M,V)


	def read_file(self,filename):
		file = open(filename,'r')
		f = file.read()
		file.close()
		f = f.strip().splitlines()
		truth = []
		uncalibrated = []
		for line in f:
			line = line.strip().split('\t')
			truth.append(int(line[0]))
			uncalibrated.append(float(line[1]))
		return truth, uncalibrated

	def plot_calibrated_reliability(self,truth,uncalibrated,frac1,ir,M,V):
		bins_0_sums = [float(i+.5)/self.numbins for i in range(self.numbins)]
		bins_0_lens = [1 for i in range(self.numbins)]
		bins_1_sums = [float(i+.5)/self.numbins for i in range(self.numbins)]
		bins_1_lens = [1 for i in range(self.numbins)]
		for i in range(len(truth)):
			val = self.second_order_correction([uncalibrated[i]],M,V)[0]
			bin = self.prob2interval(val)
			if truth[i] == 0:
				bins_0_sums[bin] += val
				bins_0_lens[bin] += 1
			elif truth[i] == 1:
				bins_1_sums[bin] += val
				bins_1_lens[bin] += 1
		xvals = []
		yvals = []
		total0 = sum(bins_0_lens)
		total1 = sum(bins_1_lens)
		newNs = float(frac1)*float(total0)/(1-frac1)
		downsample_scalar = float(newNs)/total1
		for bin in range(self.numbins):
			if bins_1_lens[bin] == 1 and bins_0_lens[bin] == 1:
				pass
			else:
				x = float(bins_0_sums[bin] + downsample_scalar*bins_1_sums[bin])/(bins_0_lens[bin] + downsample_scalar*bins_1_lens[bin])
				xvals.append(x)
				y = float(downsample_scalar*bins_1_lens[bin])/(bins_0_lens[bin]+downsample_scalar*bins_1_lens[bin])
				yvals.append(y)
		plt.plot(xvals,yvals,'o-')
		plt.plot([0,1],[0,1],':k')
		plt.xlabel('Calibrated Probability')
		plt.ylabel('Fraction with Label 1')
		plt.savefig('calibration/calibrated_reliability_frac1='+str(frac1)+'.pdf')
		plt.clf()		

	def plot_uncalibrated_reliability(self,truth,uncalibrated,frac1):
		#truth,uncalibrated = self.read_file(filename)
		xtrain = [] #for IR
		ytrain = [] #for IR
		bins_0_sums = [float(i+.5)/self.numbins for i in range(self.numbins)]
		bins_0_lens = [1 for i in range(self.numbins)]
		bins_1_sums = [float(i+.5)/self.numbins for i in range(self.numbins)]
		bins_1_lens = [1 for i in range(self.numbins)]
		for i in range(len(truth)):
			val = uncalibrated[i]
			xtrain.append(val)
			bin = self.prob2interval(val)
			#print bin
			if truth[i] == 0:
				bins_0_sums[bin] += val
				bins_0_lens[bin] += 1
			elif truth[i] == 1:
				bins_1_sums[bin] += val
				bins_1_lens[bin] += 1
		xvals = []
		yvals = []
		total0 = sum(bins_0_lens)
		total1 = sum(bins_1_lens)
		newNs = float(frac1)*float(total0)/(1-frac1)
		downsample_scalar = float(newNs)/total1
		for bin in range(self.numbins):
			if bins_1_lens[bin] == 1 and bins_0_lens[bin] == 1:
				pass
			else:
				x = float(bins_0_sums[bin] + downsample_scalar*bins_1_sums[bin])/(bins_0_lens[bin] + downsample_scalar*bins_1_lens[bin])
				xvals.append(x)
				y = float(downsample_scalar*bins_1_lens[bin])/(bins_0_lens[bin]+downsample_scalar*bins_1_lens[bin])
				yvals.append(y)
		plt.plot(xvals,yvals,'o-')
		plt.plot([0,1],[0,1],':k')
		plt.xlabel('Uncalibrated Probability')
		plt.ylabel('Fraction with Label 1')
		plt.savefig('calibration/uncalibrated_reliability_frac1='+str(frac1)+'.pdf')
		plt.clf()

		for x in xtrain:
			bin = self.prob2interval(x)
			ytrain.append(yvals[bin])
		return xtrain,ytrain
 
	def prob2interval(self,prob):
		a = float(prob)/(1.0/self.numbins)
		bin = int(a)
		if bin == self.numbins:
			bin = self.numbins-1
		return bin


	def isotonic_calibration(self,xtrain,ytrain):
		ir = IR(out_of_bounds='clip')
		ir.fit(xtrain,ytrain)
		#print ir
		return ir

	def interpolate_isotonic(self,ir,delta=1e-6):
		#learn the midpoints and the values of the intervals
		x = np.arange(0,1,0.001)
		y = ir.transform(x)
		if np.isnan(y[0]):
			y[0] = 0
		startpoints = [0]
		endpoints = []
		values = [y[0]]
		for i in range(1,len(x)):
			if abs(y[i]-values[-1]) > delta:
				startpoints.append(x[i])
				values.append(y[i])
				endpoints.append(x[i-1])
		endpoints.append(1)
		midpoints = []
		for i in range(len(startpoints)):
			midpoints.append(0.5*(endpoints[i]+startpoints[i]))
		midpoints.append(1)
		values.append(1)
		return midpoints,values

	def second_order_correction(self,cal_p_vec,M,V):
		newvals = []
		for cal_p in cal_p_vec:
			i = 0
			while cal_p > M[i]:
				i += 1
			if i == 0:
				newvals.append(0)
			else:
				leftpos = M[i-1]
				rightpos = M[i]
				leftval = V[i-1]
				rightval = V[i]
				pos = float(cal_p-leftpos)/(rightpos-leftpos)
				newval = leftval + (rightval-leftval)*pos
				newvals.append(newval)
		return newvals

#if __name__ == '__main__':
def main():
	import argparse
	parser = argparse.ArgumentParser()	
	parser.add_argument('--frac1',action='store',dest='frac1',default=0.1) #training set makeup: what percentage of training set is from label 1 (sweep)? required for both --learn and --apply
	parser.add_argument('--train',action='store_true',dest='learn')
	parser.add_argument('--input_train',action='store',dest='input_train') #file with training data, required for --learn (2 columns, no headers. first column is true values 0=neutral, 1=sweep. second column is uncalibrated probabilities)
	parser.add_argument('--apply',action='store_true',dest='apply')
	parser.add_argument('--input_apply',action='store',dest='input_apply') #file with any number of tab-delimited columns, with header line. One header line must be "uncalibrated". output will make a new file that adds an additional column called "calibrated"
	args = parser.parse_args()
	C = SmoothedIsotonicCalibration()

	if args.learn:
		C.learn_calibration(float(args.frac1),args.input_train)
	elif args.apply:
		C.apply_calibration(float(args.frac1),args.input_apply)



from __future__ import print_function
from builtins import str
from builtins import range
from builtins import object
from scipy.stats import norm
import math,sys,argparse, pickle, os
sys.path.append(os.getcwd())
import matplotlib
matplotlib.use('Agg')
from matplotlib import pyplot as plt
import numpy as np
from sklearn import mixture
#from matplotlib.mlab import bivariate_normal
from scipy.stats import multivariate_normal
from scipy.stats import norm
import matplotlib.cm as cm
import warnings
import matplotlib.cbook


class AODE_train(object):
    '''
    Given directory with component_stats.txt, scenarios.txt, and subdirectory simulations/
    containing subdirectories for each classification scenario with training examples, choose
    number of Gaussian mixture components for all distributions based on BIC minima, then
    save all necessary parameters in AODE_params/ and illustrations of learned distributions
    in component_statistic_distributions/
    '''

    def __init__(self,args):
        self.retrain = args.retrain



        self.path2allstats = args.path2files
        if self.path2allstats != '' and self.path2allstats[-1] != '/':
            self.path2allstats += '/'
        self.path2allstats += 'simulations/'
        self.path2files = args.path2files
        if self.path2files != '' and self.path2files[-1] != '/':
            self.path2files += '/'
        self.path2AODE = self.path2files+'AODE_params/'


        if os.path.isdir(self.path2AODE) == False:
            os.mkdir(self.path2AODE)

        file = open(self.path2files+'classes.txt','r')
        f = file.read()
        file.close()
        self.scenarios = [x.strip() for x in f.strip().splitlines()]


        file = open(self.path2files+'component_stats.txt','r')
        f = file.read()
        file.close()
        f = f.strip().splitlines()
        self.statlist = [x.strip() for x in f]

        self.minscores = [[] for i in range(len(self.statlist))]
        self.maxscores = [[] for i in range(len(self.statlist))]

        self.num2stat = {i:self.statlist[i] for i in range(len(self.statlist))}
        self.stat2num = {y:x for x,y in list(self.num2stat.items())}        

        self.colors = ['blue','red','green','purple','orange'] #would need to add more colors for 6+ scenarios
        self.colorspectra = [cm.Blues,cm.Reds,cm.Greens,cm.Purples,cm.Oranges] #would need to add more colors for 6+ scenarios

        self.component_nums_1D = [['n/a' for s in self.scenarios] for x in self.statlist]
        self.component_nums_2D = [[['n/a' for s in self.scenarios] for y in self.statlist] for x in self.statlist]

        if os.path.isdir(self.path2files+'BIC_plots/') == False:
            os.mkdir(self.path2files+'BIC_plots/')
        if os.path.isdir(self.path2files+'component_statistic_distributions/') == False:
            os.mkdir(self.path2files+'component_statistic_distributions')
            os.mkdir(self.path2files+'component_statistic_distributions/marginals')
            os.mkdir(self.path2files+'component_statistic_distributions/joints')
        self.read_in_all()


    def tuples(self,stat1,stat2,scenario,round1=False):
        if round1 == False:
            scores = pickle.load(open(self.path2AODE+stat1+'_'+stat2+'_'+scenario+'_tuples.p','rb'))

        else:
            print('learning '+scenario+' joint distributions for '+stat1+' and '+stat2+'...')
            scores = []

            for filename in os.listdir(self.path2allstats+scenario+'/'):
                if filename[0] != '.':
                    file = open(self.path2allstats+scenario+'/'+filename,'r')
                    f = file.read()
                    file.close()
                    f = f.strip().splitlines()
                    header = f[0].strip().split('\t')
                    f = f[1:]
                    for line in f:
                        line = line.strip().split('\t')
                        score1 = float(line[header.index(stat1)])
                        score2 = float(line[header.index(stat2)])
                        if score1 != -998 and score2 != -998:
                            scores.append((score1,score2))

            pickle.dump(scores,open(self.path2AODE+stat1+'_'+stat2+'_'+scenario+'_tuples.p','wb'), protocol=2)
        return np.array(scores)

    def singles(self,stat,scenario,round1=False):
        if round1 == False:
            scores = pickle.load(open(self.path2AODE+stat+'_'+scenario+'_singles.p','rb'))
        else:
            print('learning '+scenario+' marginal distributions for '+stat+'...')
            scores = []
            for filename in os.listdir(self.path2allstats+scenario+'/'):
                if filename[0] != '.':
                    file = open(self.path2allstats+scenario+'/'+filename,'r')
                    f = file.read()
                    file.close()
                    f = f.strip().splitlines()
                    header = f[0].strip().split('\t')
                    f = f[1:]
                    for line in f:
                        line = line.strip().split('\t')
                        score = float(line[header.index(stat)])
                        if score != -998:
                            scores.append([score])
                        
            pickle.dump(scores,open(self.path2AODE+stat+'_'+scenario+'_singles.p','wb'), protocol=2)
        SCORES = np.array(scores)
        minscore = min(SCORES)[0]
        maxscore = max(SCORES)[0]
        #RANGE = maxscore-minscore
        self.minscores[self.stat2num[stat]].append(minscore)
        self.maxscores[self.stat2num[stat]].append(maxscore)
        return np.array(scores)

    def plot_bic(self,stat1,stat2,scenario):
        S = self.tuples(stat1,stat2,scenario)
        BICs_full = []
        for n in range(1,11):
            H = mixture.GaussianMixture(n_components=n,covariance_type='full')
            H.fit(S)
            #AICs_full.append(H.aic(S))
            BICs_full.append(H.bic(S))
        minbic = min(BICs_full)
        argminbic = BICs_full.index(minbic)+1
        print('number of components for '+scenario+': '+str(argminbic))
        plt.plot(list(range(1,11)),BICs_full,'o-',color='darkblue',ms=5,markeredgecolor='none')
        plt.plot(argminbic,minbic,'o-',color='coral',ms=5,markeredgecolor='red')
        plt.xlabel('number of Gaussian mixture components')
        plt.ylabel('BIC')
        plt.savefig(self.path2files+'BIC_plots/'+stat1+'_'+stat2+'_'+scenario+'_BIC.pdf')
        plt.clf()

        self.component_nums_2D[self.stat2num[stat1]][self.stat2num[stat2]][self.scenarios.index(scenario)] = argminbic

    def plot_bic_1D(self,stat,scenario):
        S = self.singles(stat,scenario)
        BICs = []
        for n in range(1,11):
            H = mixture.GaussianMixture(n_components=n)
            H.fit(S)
            #AICs.append(H.aic(S))
            BICs.append(H.bic(S))

        minbic = min(BICs)
        argminbic = BICs.index(minbic)+1
        print('number of components for '+scenario+': '+str(argminbic))            
        plt.xlabel('number of Gaussian mixture components')
        plt.ylabel('BIC')
        plt.plot(list(range(1,11)),BICs,'o-',color='darkblue',ms=5,markeredgecolor='none')
        plt.plot(argminbic,minbic,'o-',color='coral',ms=5,markeredgecolor='red')
        plt.savefig(self.path2files+'BIC_plots/'+stat+'_'+scenario+'_BIC.pdf')
        plt.clf()
        self.component_nums_1D[self.stat2num[stat]][self.scenarios.index(scenario)] = argminbic

    def read_in_all(self):
        for stat in self.statlist:
            for scenario in self.scenarios:
                self.singles(stat,scenario,round1=True)

        for i in range(len(self.statlist)-1):
            for j in range(i+1,len(self.statlist)):
                for scenario in self.scenarios:
                    self.tuples(self.statlist[i],self.statlist[j],scenario,round1=True)

    def run_bic(self):
        for stat in self.statlist:
            print('learning number of Gaussian mixture components for '+stat)
            for scenario in self.scenarios:
                self.plot_bic_1D(stat,scenario)

        for i in range(len(self.statlist)-1):
            for j in range(i+1,len(self.statlist)):
                print('learning number of Gaussian mixture components for joint '+self.statlist[i]+', '+self.statlist[j])
                for scenario in self.scenarios:
                    self.plot_bic(self.statlist[i],self.statlist[j],scenario)

        #write marginal component_nums file
        out = open(self.path2AODE+'marginal_component_nums','w')
        header = 'statistic\t'
        for scenario in self.scenarios:
            header += scenario+'\t'
        header = header.strip()
        out.write(header+'\n')
        for i in range(len(self.component_nums_1D)):
            line = self.num2stat[i]+'\t'
            for scenario in self.scenarios:
                line += str(self.component_nums_1D[i][self.scenarios.index(scenario)])+'\t'
            line = line.strip()
            out.write(line+'\n')
        out.close()

        #write joint component_nums file
        out = open(self.path2AODE+'joint_component_nums','w')
        header = 'stat1\tstat1\t'
        for scenario in self.scenarios:
            header += scenario+'\t'
        header = header.strip()
        out.write(header+'\n')
        for i in range(len(self.component_nums_2D)-1):
            for j in range(i+1,len(self.component_nums_2D[i])):
                line = self.num2stat[i]+'\t'+self.num2stat[j]+'\t'
                for scenario in self.scenarios:
                    line += str(self.component_nums_2D[i][j][self.scenarios.index(scenario)])+'\t'
                line = line.strip()
                out.write(line+'\n')
        out.close()

    def gmm_fit(self,stat1,stat2,scenario):
        S = self.tuples(stat1,stat2,scenario)
        G = mixture.GaussianMixture(n_components=self.component_nums_2D[self.stat2num[stat1]][self.stat2num[stat2]][self.scenarios.index(scenario)],covariance_type='full')
        G.fit(S)
        pickle.dump(G,open(self.path2AODE+stat1+'_'+stat2+'_'+scenario+'_GMMparams.p','wb'), protocol=2)
        return G

    def gmm_fit_1D(self,stat,scenario):
        S = self.singles(stat,scenario)
        G = mixture.GaussianMixture(n_components=self.component_nums_1D[self.stat2num[stat]][self.scenarios.index(scenario)])
        G.fit(S)
        pickle.dump(G,open(self.path2AODE+stat+'_'+scenario+'_1D_GMMparams.p','wb'), protocol=2)
        return G

    def plot_gmm_marginals(self,stat):
        fig = plt.figure()
        for scenario in self.scenarios:
            G = self.gmm_fit_1D(stat,scenario)
            mu = G.means_
            sigma = G.covariances_
            w = G.weights_
            minscore = min(self.minscores[self.stat2num[stat]])
            maxscore = max(self.maxscores[self.stat2num[stat]])
            x = np.linspace(minscore,maxscore,100)
            Z = 0
            for i in range(len(w)):
                Z = Z + w[i]*norm.pdf(x,mu[i],sigma[i])
            plt.plot(x,Z[0],self.colors[self.scenarios.index(scenario)%len(self.colors)])
        plt.xlabel(stat)
        plt.ylabel('frequency')
        plt.legend(self.scenarios)
        plt.savefig(self.path2files+'component_statistic_distributions/marginals/'+stat+'_marginal.pdf')
        plt.clf()

    def plot_gmm_contour(self,stat1,stat2):

        fig = plt.figure()
        legendlocations = {x:[] for x in self.scenarios}
        for scenario in self.scenarios:
            G = self.gmm_fit(stat1,stat2,scenario)
            Cov = G.covariances_
            mu = G.means_
            sigma = G.covariances_
            legendlocations[scenario] = [mu[0][0]+.2*sigma[0][0][0],mu[0][1]+.2*sigma[0][1][1]]            
            w = G.weights_
            minscore1 = min(self.minscores[self.stat2num[stat1]])
            maxscore1 = max(self.maxscores[self.stat2num[stat1]])
            minscore2 = min(self.minscores[self.stat2num[stat2]])
            maxscore2 = max(self.maxscores[self.stat2num[stat2]])                            
            x = np.linspace(minscore1,maxscore1,100)
            y = np.linspace(minscore2,maxscore2,100)
            X,Y = np.meshgrid(x,y)
            Z = 0
            for i in range(len(w)):
                Z = Z + w[i]*self.bivariate_normal(X,Y, mux=mu[i][0], muy=mu[i][1], sigmax=math.sqrt(sigma[i][0][0]), sigmay=math.sqrt(sigma[i][1][1]), sigmaxy=sigma[i][0][1])
            C = plt.contour(X,Y,Z,10,cmap=self.colorspectra[self.scenarios.index(scenario)%len(self.colorspectra)])

        plt.xlabel(stat1)
        plt.ylabel(stat2)
        for scenario in self.scenarios:
            plt.text(legendlocations[scenario][0],legendlocations[scenario][1],scenario,color=self.colors[self.scenarios.index(scenario)%len(self.colors)])
        plt.savefig(self.path2files+'component_statistic_distributions/joints/'+stat1+'_'+stat2+'_joint.pdf')
        plt.clf()

    def bivariate_normal(self, X, Y, sigmax=1.0, sigmay=1.0, mux=0.0, muy=0.0, sigmaxy=0.0):
        """
        redefining bivariate_normal() since it is deprecated in newer versions of matplotlib

        Bivariate Gaussian distribution for equal shape *X*, *Y*.
        See `bivariate normal
        <http://mathworld.wolfram.com/BivariateNormalDistribution.html>`_
        at mathworld.

        Source: https://stackoverflow.com/questions/64780530/attributeerror-module-matplotlib-mlab-has-no-attribute-bivariate-normal 
        """
        Xmu = X-mux
        Ymu = Y-muy

        rho = sigmaxy/(sigmax*sigmay)
        z = Xmu**2/sigmax**2 + Ymu**2/sigmay**2 - 2*rho*Xmu*Ymu/(sigmax*sigmay)
        denom = 2*np.pi*sigmax*sigmay*np.sqrt(1-rho**2)
        return np.exp(-z/(2*(1-rho**2))) / denom

    def plot_contours(self):
        for i in range(len(self.statlist)-1):
            for j in range(i+1,len(self.statlist)):
                stat1 = self.statlist[i]
                stat2 = self.statlist[j]
                self.plot_gmm_contour(stat1,stat2)
        for stat in self.statlist:
            self.plot_gmm_marginals(stat)

    def retrain_classifier(self):
        file = open(self.path2AODE+'marginal_component_nums','r')
        f = file.read()
        file.close()
        f = f.strip().splitlines()[1:]
        for i in range(len(f)):
            f[i] = f[i].strip().split('\t')
            stat = f[i][0]
            for j in range(len(self.scenarios)):
                self.component_nums_1D[self.stat2num[stat]][j] = int(f[i][j+1])

        file = open(self.path2AODE+'joint_component_nums','r')
        f = file.read()
        file.close()
        f = f.strip().splitlines()[1:]
        for i in range(len(f)):
            f[i] = f[i].strip().split('\t')
            stat1 = f[i][0]
            stat2 = f[i][1]
            for j in range(len(self.scenarios)):
                self.component_nums_2D[self.stat2num[stat1]][self.stat2num[stat2]][j] = int(f[i][j+2])

        self.plot_contours()

#if __name__ == '__main__':
def main():
    
    #suppress matplotlib deprecation warnings
    #warnings.filterwarnings("ignore", category=matplotlib.cbook.mplDeprecation)

    parser = argparse.ArgumentParser()
    parser.add_argument('--path',action='store',dest='path2files',default='') #path to all input files (simulations in a 'simulations' directory, and compstats, scenarios files)
    parser.add_argument('--retrain',action='store_true',dest='retrain')

    args = parser.parse_args()
    A = AODE_train(args)

    if args.retrain:
        A.retrain_classifier()
        print('Training complete. Run the command swifr_test with --path2trained '+args.path2files)
    else:
        A.run_bic()
        A.plot_contours()
        print('Training complete. Run the command swifr_test with --path2trained '+args.path2files)
